import os
import stat
import shutil
import errno

def rmtree(path):
  """Delete a path recursively.

  Like shutil.rmtree, but supporting the case that some files or folder
  might have been marked read only.  """
  def chmod_retry(func, failed_path, exc_info):
    """Make sure the directories are executable and writable.
    """
    e = exc_info[1]
    if isinstance(e, OSError):
      if e.errno == errno.ENOENT:
        # because we are calling again rmtree on listdir errors, this path might
        # have been already deleted by the recursive call to rmtree.
        return
      if e.errno == errno.EACCES:
        if func is os.listdir:
          os.chmod(failed_path, 0o700)
          # corner case to handle errors in listing directories.
          # https://bugs.python.org/issue8523
          return shutil.rmtree(failed_path, onerror=chmod_retry)
        # If parent directory is not writable, we still cannot delete the file.
        # But make sure not to change the parent of the folder we are deleting.
        if failed_path != path:
          os.chmod(os.path.dirname(failed_path), 0o700)
          return func(failed_path)
    raise

  shutil.rmtree(path, onerror=chmod_retry)

def createFolder(folder, clean=False):
  if os.path.exists(folder):
    if not clean:
      return
    rmtree(folder)
  os.mkdir(folder)

def deunicodeData(data):
  if isinstance(data, list):
    return map(deunicodeData, data)
  if isinstance(data, unicode):
    return data.encode('utf8')
  if isinstance(data, dict):
    return {deunicodeData(key): deunicodeData(value)
            for key, value in data.iteritems()}
  return data



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import errno
import os
import glob
import shutil
import string
import random
from .Utils import createFolder

class SlapOSInstance(object):
  """
  Base of an software instance,
  store variables used during software installation
  """
  def __init__(self, working_directory):
    self.retry_software_count = 0
    self.retry = False
    self.working_directory = working_directory


class NodeTestSuite(SlapOSInstance):
  """
  
  """
  def __init__(self, reference, working_directory):
    d = os.path.join(working_directory, reference)
    super(NodeTestSuite, self).__init__(d)
    self.reference = reference
    self.cluster_configuration = {}
    self.test_suite_directory = os.path.join(d, 'test_suite')
    self.custom_profile_path = os.path.join(d, 'software.cfg')
    createFolder(d)

  def edit(self, **kw):
    self.__dict__.update(**kw)
    for vcs_repository in kw.get('vcs_repository_list', ()):
        buildout_section_id = vcs_repository.get('buildout_section_id')
        repository_id = buildout_section_id or \
                        vcs_repository.get('url').split('/')[-1].split('.')[0]
        repository_path = os.path.join(self.working_directory,repository_id)
        vcs_repository['repository_id'] = repository_id
        vcs_repository['repository_path'] = repository_path

  def createSuiteLog(self):
    # /srv/slapgrid/slappartXX/srv/var/log/testnode/az-D27KqX7FxJ/suite.log
    alphabets = string.digits + string.letters
    while 1:
      log_folder_name = '%s-%s' % (self.reference,
        ''.join(random.choice(alphabets) for i in xrange(10)))
      log_folder_path = os.path.join(self.log_directory, log_folder_name)
      try:
        os.makedirs(log_folder_path)
      except OSError, e:
        if e.errno != errno.EEXIST:
          raise
      else:
        break
    # XXX copy the whole content of the log viewer app
    for fname in glob.glob(os.path.join(os.path.dirname(__file__), 'js-logtail', '*')):
      shutil.copy(fname, log_folder_path)
    self.suite_log_path = os.path.join(log_folder_path, 'suite.log')
    return self.suite_log_path, log_folder_name

  @property
  def revision(self):
    return ','.join('%s=%s-%s' % (
        repository[:-11] if repository.endswith('-repository') else repository,
        count, revision)
      for repository, (count, revision) in self.revision_list)



import datetime
import json
import traceback
import time
import requests
from functools import wraps
from uritemplate import expand

import slapos.slap
from slapos.slap import SoftwareProductCollection
from requests.exceptions import HTTPError
from ..taskdistribution import SAFE_RPC_EXCEPTION_LIST
from . import logger

# max time to instance changing state: 3 hour
MAX_INSTANCE_TIME = 60*60*3

SOFTWARE_PRODUCT_NAMESPACE = "product."

SOFTWARE_STATE_UNKNOWN = "SOFTWARE_STATE_UNKNOWN"
SOFTWARE_STATE_INSTALLING = "SOFTWARE_STATE_INSTALLING"
SOFTWARE_STATE_INSTALLED = "SOFTWARE_STATE_INSTALLED"
SOFTWARE_STATE_DESTROYING = "SOFTWARE_STATE_DESTROYING"

INSTANCE_STATE_UNKNOWN = "INSTANCE_STATE_UNKNOWN"
INSTANCE_STATE_STARTING = "INSTANCE_STATE_STARTING"
INSTANCE_STATE_STARTED = "started"
INSTANCE_STATE_STARTED_WITH_ERROR = "INSTANCE_STATE_STARTED_WITH_ERROR"
INSTANCE_STATE_STOPPING = "INSTANCE_STATE_STOPPING"
INSTANCE_STATE_STOPPED = "stopped"
INSTANCE_STATE_DESTROYED = "destroyed"

TESTER_STATE_INITIAL = "TESTER_STATE_INITIAL"
TESTER_STATE_NOTHING = "TESTER_STATE_NOTHING"
TESTER_STATE_SOFTWARE_INSTALLED = "TESTER_STATE_SOFTWARE_INSTALLED"
TESTER_STATE_INSTANCE_INSTALLED = "TESTER_STATE_INSTANCE_INSTALLED"
TESTER_STATE_INSTANCE_STARTED = "TESTER_STATE_INSTANCE_STARTED"
TESTER_STATE_INSTANCE_UNINSTALLED = "TESTER_STATE_INSTANCE_UNINSTALLED"


# Simple decorator to prevent raise due small
# network failures.
def retryOnNetworkFailure(func,
    _except_list = SAFE_RPC_EXCEPTION_LIST + (
      HTTPError, slapos.slap.ConnectionError),
    ):
  def wrapper(*args, **kw):
    retry_time = 64
    while True:
      try:
        return func(*args, **kw)
      except _except_list:
        traceback.print_exc()

      print 'Network failure. Retry method %s in %i seconds' % (func, retry_time)
      time.sleep(retry_time)
      retry_time = min(retry_time*1.5, 640)

  return wraps(func)(wrapper)


class SlapOSMasterCommunicator(object):
  latest_state = None

  def __init__(self, slap, slap_supply, slap_order, url):
    self.slap = slap
    self.slap_order = slap_order
    self.slap_supply = slap_supply
    self.hateoas_navigator = self.slap._hateoas_navigator

    if url is not None and \
      url.startswith(SOFTWARE_PRODUCT_NAMESPACE):

      product = SoftwareProductCollection(logger, self.slap)
      try:
        url = product.__getattr__(url[len(SOFTWARE_PRODUCT_NAMESPACE):])
      except AttributeError as e:
        logger.warning('Error on get software release: %s ', e.message)

    self.url = url  

  @retryOnNetworkFailure
  def _supply(self, state="available"):
    if self.computer_guid is None:
      logger.info('Nothing to supply for %s.', self.name)
      return None
    logger.info('Supply %s@%s', self.url, self.computer_guid)
    return self.slap_supply.supply(self.url, self.computer_guid, state)

  @retryOnNetworkFailure
  def _request(self, state, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    if instance_title is not None:
      self.name = instance_title 
    if request_kw is not None:
      if isinstance(request_kw, basestring) or \
        isinstance(request_kw, unicode):
        self.request_kw = json.loads(request_kw)
      else:
        self.request_kw = request_kw
    if self.request_kw is None:
      self.request_kw = {}
    logger.info('Request %s@%s: %s', self.url, self.name, state)
    self.latest_state = state
    return self.slap_order.request(
            software_release=self.url,
            partition_reference=self.name,
            shared=shared,
            state=state,
            **self.request_kw)

  def isInstanceRequested(self, instance_title):
    hateoas = getattr(self.slap, '_hateoas_navigator', None)
    return instance_title in hateoas.getHostingSubscriptionDict()

  @retryOnNetworkFailure
  def _hateoas_getComputer(self, reference):
    root_document = self.hateoas_navigator.getRootDocument()
    search_url = root_document["_links"]['raw_search']['href']

    getter_link = expand(search_url, { 
      "query": "reference:%s AND portal_type:Computer" % reference, 
      "select_list": ["relative_url"],
      "limit": 1})

    result = self.hateoas_navigator.GET(getter_link)
    content_list = json.loads(result)['_embedded']['contents']

    if len(content_list) == 0:
      raise Exception('No Computer found.')

    computer_relative_url = content_list[0]["relative_url"]
    
    getter_url = self.hateoas_navigator.getDocumentAndHateoas(
      computer_relative_url)

    return json.loads(self.hateoas_navigator.GET(getter_url))
    
 
  @retryOnNetworkFailure
  def getSoftwareInstallationList(self, computer_guid=None):
    # XXX Move me to slap.py API
    computer = self._hateoas_getComputer(computer_guid) if computer_guid else self._hateoas_getComputer(self.computer_guid)

    # Not a list ?
    action = computer['_links']['action_object_slap']

    if action.get('title') == 'getHateoasSoftwareInstallationList':
      getter_link = action['href']
    else:
      raise Exception('No Link found found.')

    result = self.hateoas_navigator.GET(getter_link)
    return json.loads(result)['_links']['content']


  @retryOnNetworkFailure
  def getSoftwareInstallationNews(self, computer_guid=None):
    getter_link = None
    for si in self.getSoftwareInstallationList(computer_guid):
      if si["title"] == self.url:
        getter_link = si["href"]
        break

    # We could not find the document, so it is probably too soon.
    if getter_link is None:
      return ""
    
    result = self.hateoas_navigator.GET(getter_link)
    action_object_slap_list = json.loads(result)['_links']['action_object_slap']

    for action in action_object_slap_list:
      if action.get('title') == 'getHateoasNews':
        getter_link = action['href']
        break
    else:
      raise Exception('getHateoasNews not found.')

    result = self.hateoas_navigator.GET(getter_link)
    if len(json.loads(result)['news']) > 0:
      return json.loads(result)['news'][0]["text"]
    return ""

  @retryOnNetworkFailure
  def getInstanceUrlList(self):
    hosting_subscription_dict = self.hateoas_navigator._hateoas_getHostingSubscriptionDict()
    # Don't store hosting subscription url. It changes from time to time.
    hosting_subscription_url = None
    for hs in hosting_subscription_dict:
      if hs['title'] == self.name:
        hosting_subscription_url = hs['href']
        break

    if hosting_subscription_url is None:
      return None

    return self.hateoas_navigator.getHateoasInstanceList(
            hosting_subscription_url)

  @retryOnNetworkFailure
  def getNewsFromInstance(self, url):
    result = self.hateoas_navigator.GET(url)
    result = json.loads(result)
    if result['_links'].get('action_object_slap', None) is None:
      return None

    object_link = self.hateoas_navigator.hateoasGetLinkFromLinks(
       result['_links']['action_object_slap'], 'getHateoasNews')
    
    result = self.hateoas_navigator.GET(object_link)
    return json.loads(result)['news']

  @retryOnNetworkFailure
  def getInformationFromInstance(self, url):
    result = self.hateoas_navigator.GET(url)
    result = json.loads(result)
    if result['_links'].get('action_object_slap', None) is None:
      print result['links']
      return None

    object_link = self.hateoas_navigator.hateoasGetLinkFromLinks(
       result['_links']['action_object_slap'], 'getHateoasInformation')

    result = self.hateoas_navigator.GET(object_link)
    return json.loads(result)

  def _getSoftwareState(self, computer_guid=None):
    if self.computer_guid is None:
      return SOFTWARE_STATE_INSTALLED

    message = self.getSoftwareInstallationNews(computer_guid)
    logger.info(message)
    if message.startswith("#error no data found"):
      return SOFTWARE_STATE_UNKNOWN

    if message.startswith('#access software release'):
      return SOFTWARE_STATE_INSTALLED

    if message.startswith('#error'):
      return SOFTWARE_STATE_INSTALLING

    return SOFTWARE_STATE_UNKNOWN

  @retryOnNetworkFailure
  def getRSSEntryFromMonitoring(self, base_url):
    if base_url is None:
      return {}

    feed_url = base_url + '/monitor-public/rssfeed.html'
    d = feedparser.parse(feed_url)

    if len(d.entries) > 0:
      return {"date": d.entries[0].published,
              "message": d.entries[0].description,
              "title" : d.entries[0].title}
    return {}

  @retryOnNetworkFailure
  def _getInstanceState(self):
    latest_state = self.latest_state
    logger.info('latest_state = %r', latest_state)

    if latest_state is None:
      return INSTANCE_STATE_UNKNOWN

    message_list = []
    try:
      for instance in self.getInstanceUrlList():
        news = self.getNewsFromInstance(instance["href"])
        information = self.getInformationFromInstance(instance["href"])
        state = INSTANCE_STATE_UNKNOWN
        monitor_information_dict = {}

        info_created_at = "-1"
        is_slave = information['slave']
        if is_slave:
          if (information["connection_dict"]) > 0:
            state =  INSTANCE_STATE_STARTED
        else:
          # not slave
          instance_state = news[0]
          if instance_state.get('created_at', '-1') != "-1":
            # the following does NOT take TZ into account
            created_at = datetime.datetime.strptime(instance_state['created_at'], 
              '%a, %d %b %Y %H:%M:%S %Z')
            gmt_now = datetime.datetime(*time.gmtime()[:6])

            info_created_at = '%s (%d)' % (
               instance_state['created_at'], (gmt_now - created_at).seconds)

            if instance_state['text'].startswith('#access'):
              state =  INSTANCE_STATE_STARTED

            if instance_state['text'].startswith('#access Instance correctly stopped'):
              state =  INSTANCE_STATE_STOPPED

            if instance_state['text'].startswith('#destroy'):
              state = INSTANCE_STATE_DESTROYED

            if instance_state['text'].startswith('#error'):
              state = INSTANCE_STATE_STARTED_WITH_ERROR

        if state == INSTANCE_STATE_STARTED_WITH_ERROR:
          # search for monitor url
          monitor_v6_url = information["connection_dict"].get("monitor_v6_url")
          try:
            monitor_information_dict = self.getRSSEntryFromMonitoring(monitor_v6_url)
          except Exception:
            logger.exception('Unable to download promises for: %s', instance["title"])
            logger.error(traceback.format_exc())
            monitor_information_dict = {"message": "Unable to download"}

        message_list.append({
          'title': instance["title"],
          'slave': is_slave,
          'news': news[0],
          'information': information,
          'monitor': monitor_information_dict,
          'state': state
        })

    except slapos.slap.ServerError:
      logger.error('Got an error requesting partition for its state')
      return INSTANCE_STATE_UNKNOWN
    except Exception:
      logger.error("ERROR getting instance state")
      return INSTANCE_STATE_UNKNOWN

    started = 0
    stopped = 0
    self.message_history.append(message_list)
    for instance in message_list:
      if not instance['slave'] and \
        instance['state'] in (INSTANCE_STATE_UNKNOWN, INSTANCE_STATE_STARTED_WITH_ERROR):
        return instance['state']
      elif not instance['slave'] and instance['state'] == INSTANCE_STATE_STARTED:
        started = 1
      elif not instance['slave'] and instance['state'] == INSTANCE_STATE_STOPPED:
        stopped = 1

      if instance['slave'] and instance['state'] == INSTANCE_STATE_UNKNOWN:
        return instance['state']

    if started and stopped:
      return INSTANCE_STATE_STOPPED
      return INSTANCE_STATE_UNKNOWN

    if started:
      return INSTANCE_STATE_STARTED

    if stopped:
      return INSTANCE_STATE_STOPPED

  @retryOnNetworkFailure
  def _waitInstance(self, instance_title, state, max_time=MAX_INSTANCE_TIME):
    """
    Wait for 'max_time' an instance specific state
    """
    logger.info("Waiting for instance state: %s", state)
    start_time = time.time()
    while (not self._getInstanceState() == state
           and (max_time > (time.time()-start_time))):
      logger.info("Instance(s) not in %s state yet.", state)
      logger.info("Current state: %s", self._getInstanceState())
      time.sleep(15)
    if (time.time()-start_time) > max_time:
      error_message = "Instance '%s' not '%s' after %s seconds" %(instance_title, state, str(time.time()-start_time))
      return {'error_message' : error_message}
    logger.info("Instance correctly '%s' after %s seconds.", state, time.time() - start_time)
    return {'error_message' : None}

class SlapOSTester(SlapOSMasterCommunicator):

  def __init__(self,
              name,
              slap,
              slap_order,
              slap_supply,
              url, # software release url
              computer_guid=None, # computer for supply if desired
              request_kw=None
          ):
    super(SlapOSTester, self).__init__(
      slap, slap_supply, slap_order, url)

    self.name = name
    self.computer_guid = computer_guid

    if isinstance(request_kw, str) or \
      isinstance(request_kw, unicode):
      self.request_kw = json.loads(request_kw)
    else:
      self.request_kw = request_kw
    self.message_history = []

  def getInfo(self):
    info = ""
    info += "Software Release URL: %s\n" % (self.url)
    if self.computer_guid is not None:
      info += "Supply requested on: %s\n" % (self.computer_guid)
    info += "Instance Requested (Parameters): %s\n" % self.request_kw
    return info

  def supply(self, software_path=None, computer_guid=None, state="available"):
    if software_path is not None:
      self.url = software_path
    if computer_guid is not None:
      self.computer_guid = computer_guid
    self._supply(state)

  def requestInstanceStart(self, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    self.instance = self._request(INSTANCE_STATE_STARTED, instance_title, request_kw, shared, software_type)

  def requestInstanceStop(self, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    self.instance = self._request(INSTANCE_STATE_STOPPED, instance_title, request_kw, shared, software_type)

  def requestInstanceDestroy(self, instance_title=None, request_kw=None, shared=False):
    if not instance_title:
      instance_title = self.name
    self.destroyInstance(instance_title)

  def waitInstanceStarted(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_STARTED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      logger.error("Instance '%s' will be stopped and test aborted.", instance_title)
      self.requestInstanceStop()
      time.sleep(60)
      raise ValueError(error_message)

  def waitInstanceStopped(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_STOPPED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      raise ValueError(error_message)

  def waitInstanceDestroyed(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_DESTROYED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      raise ValueError(error_message)

  def getMasterFrontendDict(self):
    def getInstanceGuid():
      try:
        return self.instance.getInstanceGuid()
      except:
        return None
    frontend_master_ipv6 = None
    instance_guid = None
    for instance in self.getInstanceUrlList():
      if instance["title"] == "Monitor Frontend apache-frontend-1":
        try:
          information = self.getInformationFromInstance(instance["href"])
          frontend_master_ipv6 = information['parameter_dict']['url']
        except Exception as e:
          pass
    start_time = time.time()
    while not getInstanceGuid() and time.time()-start_time < 60*5:
      sleep(60)
    return {'instance_guid' : getInstanceGuid(), 'frontend_master_ipv6' : frontend_master_ipv6}

  # XXX TODO
  # In the future, this should allow customization so each project to be tested parses its own information,
  # probably in the test suite definition class
  def getInstanceUrlDict(self):
    frontend_url_list = []
    for instance in self.getInstanceUrlList():
      information = self.getInformationFromInstance(instance["href"])
      if "frontend-" in instance["title"]:
        try:
          frontend = [instance["title"].replace("frontend-", ""),
                      information["connection_dict"]["secure_access"]]
          frontend_url_list.append(frontend)
        except Exception as e:
          logger.info("Frontend url not generated yet for instance: " + instance["title"])
          pass
      if instance["title"] == self.name:
        try:
          connection_json = json.loads(information["connection_dict"]["_"])
          user = connection_json["inituser-login"]
          password = connection_json["inituser-password"]
        except Exception as e:
          raise ValueError("user and password not found in connection parameters. Error while instantiating?")
    return {'user' : user, 'password' : password, 'frontend-url-list' : frontend_url_list }

  def destroyInstance(self, instance_title):
    self.name = instance_title
    if self.getInstanceUrlList():
      for instance in self.getInstanceUrlList():
        if instance["title"] != instance_title:
          self._request(INSTANCE_STATE_DESTROYED, instance["title"])
        else:
          root_instance = instance
      logger.info("Going to destroy root partition: " + str(instance_title))
      self._request(INSTANCE_STATE_DESTROYED, root_instance["title"])
    else:
      logger.info("Instance not found")

class SoftwareReleaseTester(SlapOSTester):
  deadline = None

  def __init__(self,
              name,
              slap,
              slap_order,
              slap_supply,
              url, # software release url
              computer_guid=None, # computer for supply if desired
              request_kw=None,
              software_timeout=3600,
              instance_timeout=3600,
          ):
    super(SoftwareReleaseTester, self).__init__(
      name, slap, slap_order, slap_supply, url, computer_guid, request_kw)

    self.state = TESTER_STATE_INITIAL
    self.transition_dict = {
      # step function
      # delay
      # next_state
      # software_state
      # instance_state
      TESTER_STATE_INITIAL: (
        lambda t: None,
        None,
        TESTER_STATE_NOTHING,
        None,
        None,
      ),
      TESTER_STATE_NOTHING: (
        lambda t: t._supply("available"),
        int(software_timeout),
        request_kw is None and TESTER_STATE_INSTANCE_UNINSTALLED or \
            TESTER_STATE_SOFTWARE_INSTALLED,
        SOFTWARE_STATE_INSTALLED,
        None,
      ),
      TESTER_STATE_SOFTWARE_INSTALLED: (
        lambda t: t._request(INSTANCE_STATE_STARTED),
        int(instance_timeout),
        TESTER_STATE_INSTANCE_STARTED,
        None,
        INSTANCE_STATE_STARTED,
      ),
      TESTER_STATE_INSTANCE_STARTED: (
        lambda t: t._request(INSTANCE_STATE_DESTROYED),
        int(1200),
        TESTER_STATE_INSTANCE_UNINSTALLED,
        None,
        INSTANCE_STATE_STOPPED,
      ),
      TESTER_STATE_INSTANCE_UNINSTALLED: (
        lambda t: t._supply(INSTANCE_STATE_DESTROYED),
        int(1200),
        None,
        None,
        None,
      ),
     }

  def __repr__(self):
      deadline = self.deadline
      if deadline is not None:
          deadline -= time.time()
          deadline = '+%is' % (deadline, )
      return '<%s(state=%s, deadline=%s) at %x>' % (
          self.__class__.__name__, self.state, deadline, id(self))

  def getFormatedLastMessage(self):
    if len(self.message_history) == 0:
      return "No message"

    summary = "Summary about the test. Instance List and Status:\n"
    message = "Last information about the tester:\n"
    if self.message_history[-1] is not None:
      message_list = self.message_history[-1]
      for entry in message_list:
        summary += "%s %s -> %s\n" % (
          entry['title'], entry["slave"] and "(slave)" or "", entry['state'])
        for prop in entry:
          if prop != "information":
            message += "%s = %s\n" % (prop, json.dumps(entry[prop], indent=2))
          
        message += "=== connection_dict === \n%s\n" % (
          json.dumps(entry["information"]["connection_dict"], indent=2))
        message += "\n"
        message += "=== parameter_dict === \n%s\n" % (
          json.dumps(entry["information"]["parameter_dict"], indent=2))
        message += "\n"
      message += "="*79
      message += "\n\n\n"
 
    return summary + message

  @retryOnNetworkFailure
  def teardown(self):
    """
    Interrupt a running test sequence, putting it in idle state.
    """
    logger.info('Invoking TearDown for %s@%s', self.url, self.name)
    if self.request_kw is not None:
       self._request(INSTANCE_STATE_DESTROYED)
    if self.computer_guid is not None:
      self._supply(INSTANCE_STATE_DESTROYED)
    self.state = TESTER_STATE_INSTANCE_UNINSTALLED

  def tic(self, now):
    """
    Check for missed deadlines (-> test failure), conditions for moving to
    next state, and actually moving to next state (executing its payload).
    """
    logger.debug('TIC')
    deadline = self.deadline

    if deadline < now and deadline is not None:
      raise TestTimeout(self.state)

    _, _, next_state, software_state, instance_state = self.transition_dict[
        self.state]

    if (software_state is None or
          software_state == self._getSoftwareState()) and (
          instance_state is None or
          instance_state == self._getInstanceState()):

      logger.debug('Going to state %s (%r)', next_state, instance_state)
      if next_state is None:
        return None

      self.state = next_state
      stepfunc, delay, _, _, _ = self.transition_dict[next_state]
      self.deadline = now + delay
      stepfunc(self)
    return self.deadline



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import threading
import time
import requests

class TestThread(threading.Thread):

  def __init__(self, process_manager, command, log, env={}):
    threading.Thread.__init__(self)
    self.process_manager = process_manager
    self.command = command
    self.log = log
    self.env = env
  
  def run(self):
    self.process_manager.spawn(*self.command, **self.env)

class TestMetricThread(threading.Thread):

  def __init__(self, metric_url, log, stop_event, interval=60):
    threading.Thread.__init__(self)
    self.metric_url = metric_url
    self.log = log
    self.interval = interval
    self.stop_event = stop_event
    self.metric_list = []
    self.error_message = None

  def run(self):
    while(not self.stop_event.is_set()):
      self.stop_event.wait(-time.time() % self.interval)
      try:
        response = requests.get(self.metric_url, timeout=60)
        if response.status_code == 200:
          self.metric_list.append(response.text)
        else:
          self.log("Error getting mettrics: response.status: %s - response.text: %s" % (str(response.status_code), str(response.text)))
          self.error_message = "response.status: %s - response.text: %s" % (str(response.status_code), str(response.text))
      except Exception as e:
        self.log("Exception while getting metrics: %s" % (str(e)))
        self.error_message = "Exception while getting metrics: %s" % (str(e))

  def getMetricList(self):
    return self.metric_list

  def getErrorMessage(self):
    return self.error_message



#!/usr/bin/env python

import argparse
import os
import shutil
import time
import sys
import multiprocessing
import signal
import errno
import json
import logging
import logging.handlers
import glob
import urlparse
import httplib
import base64
import threading
from erp5.util.benchmark.argument import ArgumentType
from erp5.util.benchmark.performance_tester import PerformanceTester
from erp5.util.benchmark.thread import TestThread, TestMetricThread
from erp5.util import taskdistribution
from erp5.util.testnode import Utils
from erp5.util.testnode.ProcessManager import SubprocessError, ProcessManager, CancellationError
import datetime

MAX_INSTALLATION_TIME = 60*50
MAX_TESTING_TIME = 60
MAX_GETTING_CONNECTION_TIME = 60*5
TEST_METRIC_TIME_INTERVAL = 60
SCALABILITY_TEST_DURATION = 10*60

SCALABILITY_LOG_FILENAME = "runScalabilityTestSuite"
LOG_FILE_PREFIX = "scalability-test"
MAX_ERRORS = 10

class DummyLogger(object):
  def __init__(self, func):
    for name in ('trace', 'debug', 'info', 'warn', 'warning', 'error',
      'critical', 'fatal'):
       setattr(self, name, func)

class ScalabilityTest(object):
  def __init__(self, data, test_result):
    self.__dict__ = {}
    self.__dict__.update(data)
    self.test_result = test_result

def doNothing(**kwargs):
  pass

def makeSuite(test_suite=None, location=None, log=doNothing, **kwargs):
  import imp
  try:
    module = imp.load_source('scalability_test', location + '__init__.py')
    suite_class = getattr(module, test_suite)
    suite = suite_class(**kwargs)
  except Exception as e:
    log("[ERROR] While making suite: " + str(e))
    raise
  return suite

def createLogger(log_path):
  log_path = os.path.join(log_path, SCALABILITY_LOG_FILENAME + ".log")
  logger_format = '%(asctime)s %(name)-13s: %(levelname)-8s %(message)s'
  formatter = logging.Formatter(logger_format)
  logging.basicConfig(level=logging.INFO,
                     format=logger_format)
  logger = logging.getLogger(SCALABILITY_LOG_FILENAME)
  file_handler = logging.handlers.RotatingFileHandler(
                  filename=log_path,
                  maxBytes=20000000, backupCount=4)
  file_handler.setFormatter(formatter)
  logger.addHandler(file_handler)
  return logger

class ScalabilityLauncher(object):
  def __init__(self):
    self.__argumentNamespace = self._parseArguments(argparse.ArgumentParser(
          description='Run benchmarking scalability suites.'))
    self.logger = createLogger(self.__argumentNamespace.log_path)
    self.log = self.logger.info
    self.users_file_original_content = []
    # Proxy to with master test_result
    portal_url = self.__argumentNamespace.test_suite_master_url
    distributor = taskdistribution.TaskDistributor(portal_url, logger=DummyLogger(self.log))
    self.log(self.__argumentNamespace.test_suite_master_url)
    self.test_result = taskdistribution.TestResultProxy(
                        distributor,
                        1.0, DummyLogger(self.log),
                        self.__argumentNamespace.test_result_path,
                        self.__argumentNamespace.node_title,
                        self.__argumentNamespace.revision
                      )
  @staticmethod
  def _addParserArguments(parser):
    # Mandatory arguments
    parser.add_argument('--instance-url',
                        metavar='INSTANCE_URL',
                        help='Main url of instance to test')

    parser.add_argument('--bootstrap-password',
                        metavar='BOOTSRAP_PASSWORD',
                        help='Bootstrap password of instance objects')

    parser.add_argument('--test-result-path',
                        metavar='TEST_RESULT_PATH',
                        help='Relative path of the test result')

    parser.add_argument('--revision',
                        metavar='REVISION',
                        help='Revision of the test_suite')

    parser.add_argument('--current-test-data',
                        metavar='CURRENT_TEST_DATA',
                        help='Data of the current test')

    parser.add_argument('--test-suite',
                        metavar='TEST_SUITE',
                        help='Name of the test suite')

    parser.add_argument('--node-title',
                        metavar='NODE_TITLE',
                        help='Title of the testnode which is running this'
                              'launcher')

    parser.add_argument('--test-suite-master-url',
                        metavar='TEST_SUITE_MASTER_URL',
                        help='Url to connect to the Master testsuite taskditributor')

    parser.add_argument('--log-path',
                        metavar='LOG_PATH',
                        help='Log Path')

    parser.add_argument('--repo-location',
                        metavar='REPO_LOCATION',
                        help='Path to repository')

    parser.add_argument('--runner-path',
                        metavar='RUNNER_PATH',
                        help='runner Path')

    parser.add_argument('--metric-url',
                        metavar='METRIC_URL',
                        help='Url to connect to instance metric generator')

  @staticmethod
  def _checkParsedArguments(namespace):
    return namespace

  @staticmethod
  def _parseArguments(parser):
    ScalabilityLauncher._addParserArguments(parser)
    namespace = parser.parse_args()
    ScalabilityLauncher._checkParsedArguments(namespace)
    return namespace

  def moveLogs(self, folder_name, current_test):
    file_to_move_list = glob.glob(os.path.join(self.__argumentNamespace.log_path,
                                  "%s*.csv" %LOG_FILE_PREFIX))
    file_to_move_list += glob.glob(os.path.join(self.__argumentNamespace.log_path,
                                  "%s*.log" %LOG_FILE_PREFIX))
    root_test_dir = os.path.join(self.__argumentNamespace.log_path,
                                 "scalability-test-%s/" % current_test.relative_path.split("/")[1])
    if not os.path.exists(root_test_dir):
      os.makedirs(root_test_dir)
    new_directory_path = os.path.join(root_test_dir,
                                      folder_name)
    if not os.path.exists(new_directory_path):
      os.makedirs(new_directory_path)
    for file_to_move in file_to_move_list:
      shutil.move(file_to_move, new_directory_path)
    self.log("Logs were moved from %s to %s" % (
      self.__argumentNamespace.log_path, new_directory_path))

  def getRunningTest(self):
    """
    Return a ScalabilityTest with current running test case informations
    """
    data_array = self.__argumentNamespace.current_test_data.split(',')
    data = json.dumps({"count": data_array[0], "title": data_array[1], "relative_path": data_array[2]})
    decoded_data = Utils.deunicodeData(json.loads(data))
    return ScalabilityTest(decoded_data, self.test_result)

  def clearUsersFile(self, user_file_path):
    self.log("Clearing users file: %s" % user_file_path)
    os.remove(user_file_path)
    users_file = open(user_file_path, "w")
    for line in self.users_file_original_content:
      users_file.write(line)
    users_file.close()

  def updateUsersFile(self, user_quantity, password, user_file_path):
    self.log("Updating users file: %s" % user_file_path)
    users_file = open(user_file_path, "r")
    file_content = users_file.readlines()
    self.users_file_original_content = file_content
    new_file_content = []
    for line in file_content:
      new_file_content.append(line.replace('<password>', password).replace('<user_quantity>', str(user_quantity)))
    users_file.close()
    os.remove(user_file_path)
    users_file = open(user_file_path, "w")
    for line in new_file_content:
      users_file.write(line)
    users_file.close()

  def run(self):
    self.log("Scalability Launcher started, with:")
    self.log("Test suite master url: %s" %self.__argumentNamespace.test_suite_master_url)
    self.log("Test suite: %s" %self.__argumentNamespace.test_suite)
    self.log("Test result path: %s" %self.__argumentNamespace.test_result_path)
    self.log("Revision: %s" %self.__argumentNamespace.revision)
    self.log("Node title: %s" %self.__argumentNamespace.node_title)
    self.log("Instance url: %s" %self.__argumentNamespace.instance_url)

    error_message_set, exit_status = set(), 0
    process_manager = ProcessManager()

    # Get suite informations
    suite = makeSuite(self.__argumentNamespace.test_suite, self.__argumentNamespace.repo_location, self.log)
    test_suite_list = suite.getTestList()

    try:
      current_test = self.getRunningTest()
    except Exception as e:
      error_message = "ERROR while getting current running test: " + str(e)
      self.log(error_message)
      return error_message, 1

    self.log("Test Case %s going to be run." %(current_test.title))
    # Prepare configuration
    current_test_number = int(current_test.title)
    test_duration = suite.getTestDuration(current_test_number)
    if not test_duration or test_duration == 0:
      test_duration = SCALABILITY_TEST_DURATION
    benchmarks_path = os.path.join(self.__argumentNamespace.repo_location, suite.getTestPath())
    user_file_full_path = os.path.join(self.__argumentNamespace.repo_location, suite.getUsersFilePath())
    user_file_path = os.path.split(user_file_full_path)[0]
    user_file = os.path.split(user_file_full_path)[1]
    tester_path = self.__argumentNamespace.runner_path
    user_quantity = suite.getUserQuantity(current_test_number)
    instance_url = self.__argumentNamespace.instance_url
    metric_url = self.__argumentNamespace.metric_url

    # To take metrics
    metric_thread_stop_event = threading.Event()
    metric_thread = TestMetricThread(metric_url, self.log, metric_thread_stop_event, interval=TEST_METRIC_TIME_INTERVAL)
    metric_thread.start()

    try:
      bootstrap_password = self.__argumentNamespace.bootstrap_password
      self.updateUsersFile(user_quantity, bootstrap_password, user_file_full_path + ".py")
    except Exception as e:
      self.log("ERROR while updating file: " + str(e))

    now = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M")
    log_dir = "test-%s_%s" % (current_test.title, now)

    command_list = []
    user_index = 0
    # Prepare commands
    for test_suite in test_suite_list:
      log_file_name_prefix = "%s_%s_suite_%s" %(LOG_FILE_PREFIX, current_test.title, test_suite)
      command_list.append([tester_path,
                           instance_url,
                           str(user_quantity/len(test_suite_list)),
                           test_suite,
                           '--benchmark-path-list', benchmarks_path,
                           '--users-file-path', user_file_path,
                           '--users-file', user_file,
                           '--filename-prefix', log_file_name_prefix,
                           '--report-directory', self.__argumentNamespace.log_path,
                           '--max-errors', str(MAX_ERRORS),
                           '--user-index', str(user_index),
                           "--duration", "%d"%test_duration,
                         ])
      user_index += user_quantity/len(test_suite_list)
    # Launch commands
    exec_env = os.environ.copy()
    exec_env.update({'raise_error_if_fail': False})
    for index, command in enumerate(command_list, start=1):
      test_thread = TestThread(process_manager, command, self.log, env=exec_env)
      test_thread.start()
    # Sleep
    self.log("Going to sleep for %s seconds." % str(test_duration))
    time.sleep(test_duration)
    process_manager.killPreviousRun()
    self.moveLogs(log_dir, current_test)

    self.log("Test Case %s has finished" %(current_test.title))
    metric_thread_stop_event.set()
    time.sleep(15) # wait thread to stop
    metric_list = metric_thread.getMetricList()
    test_output = suite.getScalabilityTestOutput(metric_list)
    if not test_output:
      metric_error = metric_thread.getErrorMessage()
      if metric_error:
        error_message = str(metric_error)
      else:
        error_message = "Metric thread couldn't get any metric"
      error_message_set.add(error_message)

    if error_message_set:
      self.log("error_message_set: " + str(error_message_set))

    test_details = "number of users=%d\n"\
                    "number of tests=%d\n"\
                    "tests=%s\n"\
                    "duration=%d\n"\
                    %(
                      (user_quantity/len(test_suite_list))*len(test_suite_list),
                      len(test_suite_list),
                      '_'.join(test_suite_list),
                      test_duration
                    )
    self.log("Test details: %s" % test_details)

    # Send results to master
    self.log("Connecting to master to set test results...")
    try:
      retry_time = 2.0
      proxy = taskdistribution.ServerProxy(
                  self.__argumentNamespace.test_suite_master_url,
                  allow_none=True
              ).portal_task_distribution
      test_result_line_test = taskdistribution.TestResultLineProxy(
                                proxy, retry_time, self.logger,
                                current_test.relative_path,
                                current_test.title
                              )
      if len(error_message_set):
        self.log("Test case failed.")
        error_message = "; ".join(str(error) for error in error_message_set)
        test_result_line_test.stop(error_count=1, failure_count=1,
                                   command=test_details,
                                   stdout=error_message, stderr=error_message)
      else:
        self.log("Test case finished. Output: %s" % test_output)
        test_result_line_test.stop(stdout=test_output,
                                   command=test_details,
                                   test_count=len(test_suite_list),
                                   duration=test_duration)
    except Exception as e:
      self.log("Error during communication to master: " + str(e))
      raise e

    self.log("Test Case Stopped")
    self.clearUsersFile(user_file_full_path + ".py")

    error_message_set = None
    exit_status = 0
    self.log("Scalability Launcher finished.")
    return error_message_set, exit_status

def main():
  error_message_set, exit_status = ScalabilityLauncher().run()
  sys.exit(exit_status)



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import datetime
import os
import subprocess
import sys
import time
import glob
import SlapOSControler
import SlapOSMasterCommunicator
import json
import time
import shutil
import logging
import string
import random
import urlparse
import base64
import httplib
import Utils
import requests
import slapos.slap
import cPickle as pickle
from ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from Updater import Updater
from erp5.util import taskdistribution
from erp5.util.benchmark.thread import TestThread
# for dummy slapos answer
import signal
from . import logger

# max time to generate frontend instance: 1.5 hour
MAX_FRONTEND_TIME = 60*90
# max time to register instance to slapOSMaster: 5 minutes
MAX_CREATION_INSTANCE_TIME = 60*10
# max time for a test: 20 minutes
MAX_TEST_CASE_TIME = 60*20
# max time to prepare SlapOS for testsuite (software installation, instances requests, etc.)
MAX_PREPARE_TEST_SUITE = 3600*10*1.0 # 10 hours
# max time for a test line creation: 5 minutes
MAX_CREATION_TEST_LINE = 60*10
# max time for bootstrapping an instance site
MAX_BOOTSRAPPING_TIME = 60*30
# max time to get a connection
MAX_CONNECTION_TIME = 60*5
# time to check site bootstrap
CHECK_BOOSTRAPPING_TIME = 60*2

# runner names
PERFORMANCE_RUNNER_SCRIPT = "performance_tester_erp5"
SCALABILITY_RUNNER_SCRIPT = "runScalabilityTestSuite"
REQUEST_URL_SCRIPT = "requestUrl"
SCALABILITY_TEST = "scalability_test"
TEST_SUITE_INIT = "__init__.py"

# access SR by password
TESTNODE_USER = "testnode"
HTACCESS = "/.htaccess"
HTPASSWD = "/.htpasswd"
PASSWORD_FILE = "/sr_pass"
PASSWORD_LENGTH = 10
HOSTFILE = "/hosts"
SR_DICT = "frontend_software_dict"
INSTANCE_DICT = "instances_dict"

class ScalabilityTestRunner():
  def __init__(self, testnode):
    self.testnode =  testnode
    self.slapos_controler = SlapOSControler.SlapOSControler(
                                  self.testnode.working_directory,
                                  self.testnode.config)
    # Create the slapos account configuration file and dir
    key = self.testnode.taskdistribution.getSlaposAccountKey()
    certificate = self.testnode.taskdistribution.getSlaposAccountCertificate()
    # Get Slapos Master Url
    self.slapos_url = ''
    try:
      self.slapos_url = self.testnode.taskdistribution.getSlaposUrl()
      if not self.slapos_url:
        self.slapos_url = self.testnode.config['server_url']
    except:
      self.slapos_url = self.testnode.config['server_url']
    
    # Get Slapos Master url used for api rest (using hateoas)
    self.slapos_api_rest_url = self.testnode.taskdistribution.getSlaposHateoasUrl()

    logger.info("SlapOS Master url is: %s", self.slapos_url)
    logger.info("SlapOS Master hateoas url is: %s", self.slapos_api_rest_url)
    
    self.key_path, self.cert_path, config_path = self.slapos_controler.createSlaposConfigurationFileAccount(
                                        key, certificate, self.slapos_url, self.testnode.config)
    self.slapos_communicator = None
    # Dict containing used to store which SR is not yet correctly installed.
    # looks like: {'comp_id1':'SR_urlA', 'comp_id2':'SR_urlA',..}
    self.remaining_software_installation_dict = {}
    
    # Protection to prevent installation of softwares after checking
    self.authorize_supply = True
    self.authorize_request = False
    # Used to simulate SlapOS answer (used as a queue)
    self.last_slapos_answer = []
    self.last_slapos_answer_request = []
    # Use *.host.vifib.net frontend
    self.frontend_software = 'http://git.erp5.org/gitweb/slapos.git/blob_plain/HEAD:/software/apache-frontend/software.cfg'
    self.frontend_instance_guid = 'SOFTINST-9238'
    self.exec_env = os.environ.copy()
    
  def _prepareSlapOS(self, software_path, computer_guid, create_partition=0, state="available"):
    # create_partition is kept for compatibility
    """
    A proxy to supply : Install a software on a specific node
    """
    logger.info("testnode, supply : %s %s", software_path, computer_guid)
    if self.authorize_supply :
      self.remaining_software_installation_dict[computer_guid] = software_path
      self.slapos_communicator.supply(software_path, computer_guid, state)
      return {'status_code' : 0}                                          
    else:
      raise ValueError("Too late to supply now. ('self.authorize_supply' is False)")
      return {'status_code' : 1}     

  def _generateInstanceTitle(self, test_suite_title):
    """
    Generate an instance title using various parameter
    TODO : add some verification (to don't use unexisting variables)
    """
    instance_title = "Scalability-"
    instance_title += "("+test_suite_title+")-"
    instance_title += str(self.involved_nodes_computer_guid).replace("'","")
    instance_title += "-"+str(datetime.datetime.now().isoformat())+"-"
    instance_title += "timestamp="+str(time.time())
    return instance_title

  def _generateInstanceXML(self, software_configuration,
                           test_result, test_suite, frontend_software, frontend_instance_guid=None):
    """
    Generate a complete scalability instance XML configuration
    """
    config = software_configuration.copy()
    config.update({'scalability-launcher-computer-guid':self.launcher_nodes_computer_guid[0]})
    config.update({'scalability-launcher-title':self.testnode.config['test_node_title']})
    config.update({'test-result-path':test_result.test_result_path})
    config.update({'test-suite-revision':test_result.revision})
    config.update({'test-suite':test_suite})
    config.update({'test-suite-master-url':self.testnode.config['test_suite_master_url']})
    if frontend_instance_guid:
      config["frontend"] = {"software-url": frontend_software}
      config["sla-dict"]["instance_guid=%s" % frontend_instance_guid] = ["frontend"]
    return config

  def getDictionaryFromFile(self, dict_file):
    dictionary = {}
    if os.path.isfile(dict_file):
      with open(dict_file, 'r') as file:
        dictionary = pickle.loads(file.read())
    return dictionary

  def updateDictionaryFile(self, dict_file, dictionary, key, value):
    dictionary[key] = value
    with open(dict_file, 'w') as file:
      file.write(pickle.dumps(dictionary))

  def _createInstance(self, software_path, software_configuration,
                      test_result, test_suite, frontend_software, frontend_instance_guid=None):
    """
    Create scalability instance. If there is an old instance for this testsuite, it destroys it.
    """
    if self.authorize_request:
      slappart_directory = self.testnode.config['srv_directory'].rsplit("srv", 1)[0]
      instance_dict_file = slappart_directory + "var/" + INSTANCE_DICT
      instance_dict = self.getDictionaryFromFile(instance_dict_file)
      if test_suite in instance_dict:
        self.slapos_communicator.requestInstanceDestroy(instance_dict[test_suite])
      self.updateDictionaryFile(instance_dict_file, instance_dict, test_suite, self.instance_title)
      logger.info("testnode, request : %s", self.instance_title)
      config = self._generateInstanceXML(software_configuration,
                                         test_result, test_suite, frontend_software, frontend_instance_guid)
      request_kw = {"partition_parameter_kw": {"_" : json.dumps(config)} }

      self.slapos_communicator.requestInstanceStart(self.instance_title, request_kw)
      self.authorize_request = False
      return {'status_code' : 0}                                          
    else:
      raise ValueError("Softwares release not ready yet to launch instan\
ces or already launched.")
      return {'status_code' : 1}

  def prepareSlapOSForTestNode(self, test_node_slapos=None):
    """
    We will build slapos software needed by the testnode itself,
    """
    if self.testnode.taskdistribution.isMasterTestnode(
                           self.testnode.config['test_node_title']):
      pass
    return {'status_code' : 0} 

  # Dummy slapos answering
  def _getSignal(self, signal, frame):
    logger.debug("Dummy SlapOS Master answer received.")
    self.last_slapos_answer.append(True)
  def _prepareDummySlapOSAnswer(self):
    pid = os.getpid()
    logger.info("Dummy slapOS answer enabled, send signal to %s (kill -USR1 %s)"
                " to simulate a SlapOS (positive) answer.", pid, pid)
    signal.signal(signal.SIGUSR1, self._getSignal)
  def _comeBackFromDummySlapOS(self):
    logger.info("Dummy slapOS answer disabled, please don't send more signals.")
    # use SIG_USR (kill)
    signal.signal(signal.SIGUSR1, signal.SIG_DFL)
  def simulateSlapOSAnswer(self):
    if len(self.last_slapos_answer)==0:
      return False
    else:
      return self.last_slapos_answer.pop()
  # /Dummy slapos answering
    
  def isSoftwareReleaseReady(self, software_url, computer_guid):
    """
    Return true if the specified software on the specified node is installed.
    This method should communicates with SlapOS Master.
    """
    software_state = self.slapos_communicator._getSoftwareState(computer_guid)
    logger.info("Current software state: %s", software_state)
    return software_state == SlapOSMasterCommunicator.SOFTWARE_STATE_INSTALLED

  def remainSoftwareToInstall(self):
    """
    Return True if it remains softwares to install, otherwise return False
    """
    # Remove from grid installed software entries
    for computer_guid, software_path in self.remaining_software_installation_dict.items():
      if self.isSoftwareReleaseReady(software_path, computer_guid):
        del self.remaining_software_installation_dict[computer_guid]
    # Not empty grid means that all softwares are not installed
    return len(self.remaining_software_installation_dict) > 0

  def _updateInstanceXML(self, software_configuration, instance_title,
                         test_result, test_suite, frontend_software, frontend_instance_guid=None):
    logger.info("testnode, updateInstanceXML : %s", instance_title)
    config = self._generateInstanceXML(software_configuration,
                                       test_result, test_suite, frontend_software, frontend_instance_guid)
    request_kw = {"partition_parameter_kw": {"_" : json.dumps(config)} }
    self.slapos_communicator.requestInstanceStart(instance_title, request_kw)

    logger.info("Waiting for new configuration")
    time.sleep(60*5)
    self.slapos_communicator.requestInstanceStop(instance_title, request_kw)
    self.slapos_communicator.waitInstanceStopped(instance_title)
    self.slapos_communicator.requestInstanceStart(instance_title, request_kw)

    return {'status_code' : 0} 

  def _waitInstanceCreation(self, instance_title, max_time=MAX_CREATION_INSTANCE_TIME):
    """
    Wait for 'max_time' the instance creation
    """
    logger.debug("Waiting for instance creation...")
    start_time = time.time()
    while (not self.slapos_communicator.isInstanceRequested(instance_title) \
           and (max_time > (time.time()-start_time)) ):
      logger.debug("Instance not ready yet. Sleeping 5 sec.")
      time.sleep(5)
    if (time.time()-start_time) > max_time:
      raise ValueError("Instance '%s' not found after %s seconds" %(instance_title, max_time))
    logger.debug("Instance found on slapOSMaster")

  def _initializeSlapOSConnection(self):
    """
    Initialize communication with slapos 
    """
    slap = slapos.slap.slap()
    retry = 0
    while True:
      # wait until _hateoas_navigator is loaded.
      if retry > 100:
         break
      slap.initializeConnection(self.slapos_url, 
                                self.key_path, 
                                self.cert_path, 
                                timeout=120, 
                                slapgrid_rest_uri=self.slapos_api_rest_url)
      if getattr(slap, '_hateoas_navigator', None) is None:
         retry += 1
         logger.info("Fail to load _hateoas_navigator waiting a bit and retry.")
         time.sleep(30)
      else:
         break

    if getattr(slap, '_hateoas_navigator', None) is None:
      raise ValueError("Fail to load _hateoas_navigator")

    supply = slap.registerSupply()
    order = slap.registerOpenOrder()
    return slap, supply, order

  def generateProfilePasswordAccess(self):
    software_hash_directory = self.testnode.config['slapos_binary'].rsplit("bin/slapos", 1)[0]
    apache_htpasswd = software_hash_directory + "parts/apache/bin/htpasswd"
    testsuite_directory = self.testnode.config['repository_path_list'][0].rsplit('/', 1)[0]
    htaccess_file = open(testsuite_directory + HTACCESS, "w")
    file_content = """
AuthType Basic
AuthName "Password Protected Area"
AuthUserFile "%s%s"
Require valid-user
""" % (testsuite_directory, HTPASSWD)
    htaccess_file.write(file_content)
    htaccess_file.close()
    password_path = testsuite_directory + PASSWORD_FILE
    with open(password_path, "w") as password_file:
      password = ''.join(random.choice(string.digits + string.letters) for i in xrange(PASSWORD_LENGTH))
      password_file.write(password)
    user = TESTNODE_USER
    command = [apache_htpasswd, "-bc", testsuite_directory + HTPASSWD, user, password]
    self.testnode.process_manager.spawn(*command)
    return user, password

  def createSoftwareReachableProfilePath(self, node_test_suite):
    # Create an obfuscated link to the testsuite directory
    path_to_suite = os.path.join(
                    self.testnode.config['working_directory'],
                    node_test_suite.reference)
    self.obfuscated_link_path = os.path.join(
                    self.testnode.config['software_directory'],
                    self.randomized_path)
    if ( not os.path.lexists(self.obfuscated_link_path) and
         not os.path.exists(self.obfuscated_link_path) ) :
      try :
        os.symlink(path_to_suite, self.obfuscated_link_path)
        logger.info("testnode, Symbolic link (%s->%s) created.",
                    self.obfuscated_link_path, path_to_suite)
      except Exception:
        msg = "testnode, Unable to create symbolic link to the testsuite."
        logger.exception(msg)
        raise ValueError(msg)
    logger.info("Sym link : %s %s", path_to_suite, self.obfuscated_link_path)

    user, password = self.generateProfilePasswordAccess()
    logger.info("Software Profile password: %s" % password)
    self.reachable_profile = "https://%s:%s@%s" % (user, password,
      os.path.join(urlparse.urlparse(self.testnode.config['frontend_url']).netloc,
                   "software", self.randomized_path, "software.cfg"))

  def prepareSlapOSForTestSuite(self, node_test_suite):
    """
    Install testsuite softwares
    """
    logger.debug('Preparing SlapOS for Test Suite...')
    max_time = MAX_PREPARE_TEST_SUITE
    interval_time = 60
    start_time = time.time()

    # Initialize communication with slapos
    slap, supply, order = self._initializeSlapOSConnection()

    # Only master testnode must order software installation
    if self.testnode.taskdistribution.isMasterTestnode(self.testnode.config['test_node_title']):
      # Get from ERP5 Master the configuration of the cluster for the test
      test_configuration = Utils.deunicodeData(
          json.loads(self.testnode.taskdistribution.generateConfiguration(
                     node_test_suite.test_suite_title)))
      self.involved_nodes_computer_guid = test_configuration['involved_nodes_computer_guid']
      self.launchable = test_configuration['launchable']
      self.error_message = test_configuration['error_message']
      self.randomized_path = test_configuration['randomized_path']
      if not self.launchable:
        logger.info("Test suite %s is not actually launchable"
                    " with the current cluster configuration.", node_test_suite.test_suite_title)
        logger.info("ERP5 Master indicates : %s", self.error_message)
        return {'status_code' : 1}

      configuration_list = test_configuration['configuration_list']
      node_test_suite.edit(configuration_list=configuration_list)
      self.launcher_nodes_computer_guid = test_configuration['launcher_nodes_computer_guid']
      self.instance_title = self._generateInstanceTitle(node_test_suite.test_suite_title)

      self.createSoftwareReachableProfilePath(node_test_suite)
      logger.info("Software reachable profile path is: %s", self.reachable_profile)

      # Initialize SlapOS Master Communicator
      self.slapos_communicator = SlapOSMasterCommunicator.SlapOSTester(
					self.instance_title,
					slap, 
					order, 
					supply, 
					self.reachable_profile, 
					computer_guid=self.launcher_nodes_computer_guid[0])
      # Ask for SR installation
      for computer_guid in self.involved_nodes_computer_guid:
        self._prepareSlapOS(self.reachable_profile, computer_guid) 
      # From the line below we would not supply any more softwares
      self.authorize_supply = False
      # TODO : remove the line below wich simulate an answer from slapos master
      self._prepareDummySlapOSAnswer()
      # Waiting until all softwares are installed
      while (self.remainSoftwareToInstall() 
             and (max_time > (time.time()-start_time))):
        logger.info("Master testnode is waiting for the end of "
                    "all software installation (for %ss) PID=%s.",
                    int(time.time()-start_time), os.getpid())
        time.sleep(interval_time)
      # TODO : remove the line below wich simulate an answer from slapos master
      self._comeBackFromDummySlapOS()
      if self.remainSoftwareToInstall() :
        # All softwares are not installed, however maxtime is elapsed, that's a failure.
        logger.error("All softwares are not installed.")
        return {'status_code' : 1}
      self.authorize_request = True
      logger.debug("All software installed.")
      # Launch instance
      try:
        self._createInstance(self.reachable_profile, 
                             configuration_list[0],
                             node_test_suite.test_result,
                             node_test_suite.test_suite,
                             self.frontend_software,
                             self.frontend_instance_guid)
        logger.debug("Scalability instance requested.")
        self._waitInstanceCreation(self.instance_title)
      except Exception as e:
        logger.error("Error creating instance: " + str(e))
        return {'status_code' : 1}

      return {'status_code' : 0}
    return {'status_code' : 1}

  def makeSuite(self, test_suite, location_list, **kwargs):
    import imp
    suite = None
    repo_location = None
    for location in location_list:
      try:
        module = imp.load_source(SCALABILITY_TEST, "%s/%s/%s" %(location, SCALABILITY_TEST, TEST_SUITE_INIT))
        suite_class = getattr(module, test_suite)
        suite = suite_class(**kwargs)
        repo_location = "%s/%s/" % (location, SCALABILITY_TEST)
      except:
        pass
    return suite, repo_location

  def getInstanceInformation(self, suite, count):
    logger.info("Getting instance information:")
    instance_information_time = time.time()
    instance_information = self.slapos_communicator.getInstanceUrlDict()
    while not instance_information['frontend-url-list'] and \
          time.time() - instance_information_time < MAX_FRONTEND_TIME:
      time.sleep(5*60)
      instance_information = self.slapos_communicator.getInstanceUrlDict()
    logger.info(instance_information)
    if not instance_information['frontend-url-list']:
      raise ValueError("Error getting instance information: frontend url not available")
    instance_url = suite.getScalabilityTestUrl(instance_information)
    bootstrap_url = suite.getBootstrapScalabilityTestUrl(instance_information, count)
    metric_url = suite.getScalabilityTestMetricUrl(instance_information)
    site_availability_url = suite.getSiteAvailabilityUrl(instance_information)
    return instance_url, bootstrap_url, metric_url, site_availability_url

  def bootstrapInstance(self, bootstrap_url):
    bootstrap_password = error_message = None
    if bootstrap_url:
      try:
          logger.info("Bootstrapping instance...")
          command = [self.requestUrlScript, bootstrap_url]
          result = self.testnode.process_manager.spawn(*command, **self.exec_env)
          if result['status_code']:
            error_message = "Error while bootstrapping: " + str(result['stdout'])
          else:
            response_dict = json.loads(result['stdout'])
            if response_dict["status_code"] != 0:
              error_message = "Error while bootstrapping: " + str(response_dict["error_message"])
            bootstrap_password = response_dict["password"]
          logger.info("Bootstrap password: " + str(bootstrap_password))
      except Exception as e:
        error_message = "Error while bootstrapping: " + str(e)
    return bootstrap_password, error_message

  def isInstanceReady(self, site_availability_url):
    error_message = None
    if not site_availability_url:
      return True, error_message
    start_time = time.time()
    while MAX_BOOTSRAPPING_TIME > time.time()-start_time:
      try:
        command = [self.requestUrlScript, site_availability_url]
        result = self.testnode.process_manager.spawn(*command, **self.exec_env)
        if result['status_code']:
          logger.info("Error checking site availability. Response: " + str(result['stdout']))
        if int(result['stdout']) == 0:
          logger.info("Site ready for test.")
          return True, error_message
        logger.info("Site is not available yet.")
        logger.info("[DEBUG] Site availability url response: %d" % int(result['stdout']))
        time.sleep(CHECK_BOOSTRAPPING_TIME)
      except Exception as e:
        logger.info("Error checking site availability: " + str(e))
        time.sleep(CHECK_BOOSTRAPPING_TIME)
    error_message = "Error while bootstrapping: max bootstrap time exceded."
    return False, error_message

  def runTestSuite(self, node_test_suite, portal_url):
    if not self.launchable:
      return {'status_code' : 1, 'error_message': "Current test_suite is not actually launchable." }
    configuration_list = node_test_suite.configuration_list
    test_list = range(0, len(configuration_list))
    try:
      test_result_proxy = self.testnode.taskdistribution.createTestResult(
                          node_test_suite.revision, test_list,
                          self.testnode.config['test_node_title'],
                          True, node_test_suite.test_suite_title,
                          node_test_suite.project_title)
    except Exception as e:
      return {'status_code' : 1, 'error_message': "Error in communication with master: " + str(e) }
    logger.debug("Test Result created.")
    count = 0
    error_message = None

    suite, repo_location = self.makeSuite(node_test_suite.test_suite, self.testnode.config['repository_path_list'])
    if suite == None:
      error_message = "Could not find test suite %s in the repositories" % node_test_suite.test_suite
      test_result_proxy.reportFailure(stdout=error_message)
      return {'status_code' : 1, 'error_message': error_message }

    # Each cluster configuration is tested
    for configuration in configuration_list:
      # First configuration doesn't need XML configuration update.
      if count > 0:
        logger.info("Requesting instance with configuration %d" % count)
        try:
          self._updateInstanceXML(configuration, self.instance_title,
                                  node_test_suite.test_result, node_test_suite.test_suite, self.frontend_software, self.frontend_instance_guid)
        except Exception as e:
          error_message = "Error updating instance configuration: " + str(e)
      try:
        self.slapos_communicator.waitInstanceStarted(self.instance_title)
      except Exception as e:
          error_message = "Error starting instance: " + str(e)
      if error_message:
        test_result_proxy.reportFailure(stdout=error_message)
        return {'status_code' : 1, 'error_message': error_message }
      logger.debug("INSTANCE CORRECTLY STARTED")

      # Start only the current test
      exclude_list=[x for x in test_list if x!=test_list[count]]
      try:
        test_result_line_proxy = test_result_proxy.start(exclude_list)
      except Exception as e:
        error_message = "Error in communication with master: " + str(e)
        break
      if test_result_line_proxy == None :
        error_message = "Test case already tested."
        break
      logger.info("Test case for count : %d is in a running state." % count)

      try:
        instance_url, bootstrap_url, metric_url, site_availability_url = self.getInstanceInformation(suite, count)
      except Exception as e:
        error_message = "Error getting testsuite information: " + str(e)
        break

      software_bin_directory = self.testnode.config['slapos_binary'].rsplit("slapos", 1)[0]
      self.requestUrlScript = software_bin_directory + REQUEST_URL_SCRIPT

      bootstrap_password, error_message = self.bootstrapInstance(bootstrap_url)
      if error_message: break
      instance_ready, error_message = self.isInstanceReady(site_availability_url)
      if error_message: break

      runner = software_bin_directory + PERFORMANCE_RUNNER_SCRIPT
      scalabilityRunner = software_bin_directory + SCALABILITY_RUNNER_SCRIPT
      slappart_directory = self.testnode.config['srv_directory'].rsplit("srv", 1)[0]
      log_path = slappart_directory + "var/log/"
      try:
        current_test = test_result_proxy.getRunningTestCase()
        current_test_data_json = json.loads(current_test)
        current_test_data = "%d,%s,%s" % (current_test_data_json["count"], current_test_data_json["title"], current_test_data_json["relative_path"])
      except Exception as e:
        error_message = "Error getting current test case information: " + str(e)
        break
      command = [ scalabilityRunner,
                  "--instance-url", instance_url,
                  "--bootstrap-password", bootstrap_password,
                  "--test-result-path", node_test_suite.test_result.test_result_path,
                  "--revision", node_test_suite.revision,
                  "--current-test-data", current_test_data,
                  "--node-title", self.testnode.config['test_node_title'],
                  "--test-suite-master-url", self.testnode.config["test_suite_master_url"],
                  "--test-suite", node_test_suite.test_suite,
                  "--runner-path", runner,
                  "--repo-location", repo_location,
                  "--log-path", log_path,
                  "--metric-url", metric_url
                ]
      logger.info("Running test case...")
      test_thread = TestThread(self.testnode.process_manager, command, logger.info, env=self.exec_env)
      test_thread.start()

      # Wait for test case ending
      test_case_start_time = time.time()
      try:
        while test_result_line_proxy.isTestCaseAlive() and \
              test_result_proxy.isAlive() and \
              time.time() - test_case_start_time < MAX_TEST_CASE_TIME:
          time.sleep(60)
        if test_result_line_proxy.isTestCaseAlive():
          error_message = "Test case during for %s seconds, too long. (max: %s seconds). Test failure." \
                              %(str(time.time() - test_case_start_time), MAX_TEST_CASE_TIME)
          break
        logger.info("[DEBUG] Test case for current configuration finished")
        count += 1
        # Test cancelled, finished or in an undeterminate state.
        if not test_result_proxy.isAlive():
          logger.info("[DEBUG] Test case finished")
          if count == len(configuration_list):
            break
          # Cancelled or in an undeterminate state.
          error_message = "Test case cancelled or undeterminate state."
          break
      except Exception as e:
        error_message = "Error in communication with master: " + str(e)
        break

    # Remove installed software
    self.authorize_supply = True
    logger.info("Removing installed software")
    for computer_guid in self.involved_nodes_computer_guid:
      self._prepareSlapOS(self.reachable_profile, computer_guid, state="destroyed")

    # If error appears then that's a test failure.    
    if error_message:
      logger.info("There was an error during the testsuite run: " + str(error_message))
      try:
        test_result_line_proxy.stop(error_count=1, failure_count=1,
                                    stdout=error_message, stderr=error_message)
        test_result_proxy.reportFailure(stdout=error_message)
      except Exception as e:
        error_message = "Error in communication with master: " + str(e)
      logger.debug("Test Failed.")
      return {'status_code' : 1, 'error_message': error_message} 
    # Test is finished.
    logger.info("Test finished.")

    return {'status_code' : 0}

  def getRelativePathUsage(self):
    """
    Used by the method testnode.constructProfile() to know
    if the software.cfg have to use relative path or not.
    """
    return True



#!/usr/bin/env python

import sys
import requests

class Connection(object):
  def get(self, url):
    error_message_set = set()
    result = None
    try:
      response = requests.get(url, timeout=300)
      if response.status_code != 200:
        raise ValueError("response status %d" % response.status_code)
      result = response.text
    except Exception as e:
      error_message_set.add("Error: %s" % str(e))
      return result, error_message_set
    return result, error_message_set

def main():
  exit_status = 0
  error_message_set = set()
  if len(sys.argv) != 2:
    error_message_set.add("Error in parameters, please run : requestUrl [URL]")
  else:
    result, error_message_set = Connection().get(sys.argv[1])
  if error_message_set:
    exit_status = 1
    for error in error_message_set:
      print error
  elif result:
    print result
  sys.exit(exit_status)
